import greenfoot.*;
import java.util.*;

/**
 * Noughts and Crosses game logic + AI
 * 
 * @author na200
 * @version 2011.04.16
 */
public class Game extends Actor 
{
    public static final int SIZE = 3; //Grid size currently set to (3x3) but can be changed, 
                                      //in which case performance of AI will have to be improved 
                                      //(e.g. through things like alpha-beta pruning, etc)
    public static final char EMPTY_CELL = ' ';
    public static final char NOUGHT = 'o';
    public static final char CROSS = 'x';
    private Board boardGUI;
    private char[][] board;
    private GameStatus status;
    private int moveCount = 0;

    /**
     * Constructor for objects of class Game
     */
    public Game(Board boardGUI)
    {
        this.boardGUI = boardGUI;
        board = new char[SIZE][SIZE];
        newGame();
    }
    
    /**
     * Starts a new game
     */
    public void newGame()
    {
        clearBoard();
        status = GameStatus.RUNNING;
    }    
        
    /**
     * Updates game status, Handles user input and let's the computer play.
     */
    public void act()
    {
        //human-player's turn
        updateStatus();        
        if(finished()){
            gameOver();
            return;
        }
        
        //computer's turn
        addCross();        
        updateStatus();
        if(finished()){
            gameOver();
            return;
        }
    }
    
    /**
     * Adds a nought at the given position on the board.
     * Called by user through the GUI.
     * @param x X coordinate
     * @param y Y coordinate
     * @return true if successful or false otherwise
     */
    public boolean addNought(int x, int y)
    {
       if(!finished() && isValidMove(board, x, y)){
           board[x][y] = NOUGHT;
           boardGUI.addNought(x, y);
           return true;
        }
        
        return false;
    }
            
                    
    //========================================== Private ===============================================//
    
    /**
     * Determines if gameplay ended.
     */
    private boolean finished()
    {
        return status != GameStatus.RUNNING;
    }
    
    /**
     * Adds a cross at the given position - used by the computer player
     */
    private boolean addCross(int x, int y)
    {
       if(!finished() && isValidMove(board, x, y)){
           board[x][y] = CROSS;
           boardGUI.addCross(x, y);
           return true;
        }
        
        return false;
    }    
        
    /**
     * Gameover operations
     */
    private void gameOver()
    {
        Greenfoot.delay(5);
        int option = javax.swing.JOptionPane.showConfirmDialog(
            null, "Try again?", status.toString(), javax.swing.JOptionPane.YES_NO_OPTION);
        if(option == 0){
            clearBoard();
            newGame();
        }
    }
    
    /**
     * Updates game status - called during each iteration
     * Determines if game has finished, anyone has won, etc
     */
    private void updateStatus()
    {
        status = getGameStatus(board);
    }
            
    /**
     * Determines if the given position is valid to be played
     * Only empty positions on the boardGUI are considered valid
     */
    private boolean isValidMove(char[][]board, int x, int y)
    {
        return board[x][y] == EMPTY_CELL;
    }        
    
    /**
     * Determines if game finished
     * All cells occupied or one of the players won
     */
    private GameStatus getGameStatus(char[][] board)
    {
        if(gameWon(board, NOUGHT)){
            return GameStatus.PLAYER_WIN;
        }
        else if(gameWon(board, CROSS)){
            return GameStatus.COMPUTER_WIN;
        }
        else if(gameDrawn(board)){
            return GameStatus.DRAW;
        }
        else{
            return GameStatus.RUNNING;        
        }
    }

    /**
     * Determines if game is drawn, on the assumption that winning/losing situations have already been handled.
     */
    private boolean gameDrawn(char[][] board)
    {
        //all cells occupied...Draw
        for(int i = 0; i < SIZE; i++){
            for(int j = 0; j < SIZE; j++){
                if(board[j][i] == EMPTY_CELL){
                    return false;
                }
            }
        }
        return true;        
    }
    
    /**
     * Determines if the player with the given marker (e.g. 'x' or 'o') has won the game.
     */
    private boolean gameWon(char[][] board, final char MARKER)
    {
        boolean diagonalFound = true, reverseDiagonalFound = true; 
        boolean colFound, rowFound;
        
        //check columns and rows
        for(int i = 0; i < SIZE; i++)
        {
            colFound = true; rowFound = true;
            for(int j = 0; j < SIZE; j++){
                colFound = colFound && board[j][i] == MARKER;
                rowFound = rowFound && board[i][j] == MARKER;
            }
            //column or row found
            if(colFound || rowFound){
                return true;
            }
            
            //check diagonally (NOT Diagon Alley!!)...
            diagonalFound = diagonalFound && board[i][i] == MARKER;
            reverseDiagonalFound = reverseDiagonalFound && board[SIZE - 1 - i][i] == MARKER;
        }
        
        //diagonal row found
        if(diagonalFound || reverseDiagonalFound){
            return true;
        }
        
        return false;
    }
            
    /**
     * Adds a cross - used by the computer player
     */
    private void addCross()
    {
        Point point = getNextMove();
        board[point.getX()][point.getY()] = CROSS;
        boardGUI.addCross(point.getX(), point.getY());        
    }    
            
    /**
     * Determines a random move - not very intelligent.
     * Not currently in use - Used prior to the AI implementation.
     */
    private Point getRandomMove()
    {
        int x = 0, y = 0;
        do{
            x = Greenfoot.getRandomNumber(SIZE);   
            y = Greenfoot.getRandomNumber(SIZE);
        } while(!isValidMove(board, x, y));
        
        return new Point(x, y);        
    }
    
    /**
     * Clears the game board and its associated GUI
     */
    private void clearBoard()
    {
        for(int i = 0; i < SIZE; i++){
            for(int j = 0; j < SIZE; j++){
                board[j][i] = EMPTY_CELL;
            }
        }
        boardGUI.removeObjects(boardGUI.getObjects(null));
    }    
    
    /**
     * Prints board to console - useful for tesing/debugging
     */
    private void printBoard(char[][] board)
    {
        for(int i = 0; i < SIZE; i++){
            System.out.print("|");
            for(int j = 0; j < SIZE; j++){
                System.out.print(board[j][i] + "|");
            }
            System.out.println();
        }
    }    
    
    //=============================================== AI Section ============================================//
    //For details regarding the Minimax algorithm, the book 'Artificial Intelligence: A Modern Approach' 
    //by Stuart Russell et. al. is recommended.  I found the pseudocode on page-126 particularly useful.
        
    /**
     * Determines/returns next move to be played by the computer
     */
    private Point getNextMove()
    {        
        Node node = minimax(board);
        return node.getMove();
    }
    
    /**
     * Determines next best move for CROSS (i.e. computer player) via the Minimax algorithm
     */
    private Node minimax(char[][] board)
    {
        List<Node> moves = nextCandidateMoves(board, CROSS);
        
        for(Node move : moves){
            move.setScore(minimaxValue(move, false, 1));
        }
        
        return Collections.max(moves);
    }
    
    /**
     * Depth-first minimax score calculation
     * @param current Node to be determined
     * @param switchPlayer Flag used in recursive calls (in gametree generation)
     * @param depth Current tree depth (useful for testing)
     */
    private int minimaxValue(Node current, boolean switchPlayer, int depth)
    {
        //Get successor nodes of current node
        List<Node> successors = nextCandidateMoves(current.getBoard(), switchPlayer ? CROSS : NOUGHT);
        
        //Base case: 
        //We have reached a terminal node or a 'leaf' in the game-tree. Determine outcome.
        if(successors == null){
            if(gameWon(current.getBoard(), CROSS)){
                return 1;
            }
            else if(gameWon(current.getBoard(), NOUGHT)){
                return -1;
            }
            else if(gameDrawn(current.getBoard())){
                return 0;
            }
        }
        
        //Recursive case, analyse successor nodes
        //by recursively calling minimaxValue, switching players at the next level
        List<Integer> scores = new ArrayList<Integer>();
        for(Node child : successors){
            int score = minimaxValue(child, !switchPlayer, depth + 1);
            scores.add(score);
        }
        //Determine parent score from children scores
        if(switchPlayer){
            return Collections.max(scores);
        }
        else{
            return Collections.min(scores);
        }
    }
        
    /**
     * Generates list of available moves from current configuration
     * Returns null if no moves available (i.e. in case of terminal node)
     * @param current Current board configuration
     * @param MARKER marker (e.g. NOUGHT/CROSS)
     */
    private List<Node> nextCandidateMoves(char[][] current, final char MARKER)
    {
        //Terminal node. No more moves left.
        if(getGameStatus(current) != GameStatus.RUNNING){
            return null;
        }
        
        //Determine successor moves
        List<Node> moves = new LinkedList<Node>();
        for(int i = 0; i < SIZE; i++){
            for(int j = 0; j < SIZE; j++){
                if(isValidMove(current, i , j)){
                    char[][] move = getBoardCopy(current);
                    move[i][j] = MARKER;
                    moves.add(new Node(move, new Point(i, j)));
                }
            }
        }
        
        return moves;
    }
    
    /**
     * Returns a copy of the given board cofiguration
     */
    private char[][] getBoardCopy(char[][] board)
    {
        char[][] copy = new char[SIZE][SIZE];
        for(int k = 0; k < SIZE; k++){
            for(int l = 0; l < SIZE; l++){
                copy[k][l] = board[k][l];
            }
        }      
        return copy;
    }
    
    
    /**
     * Represents a gametree node
     */
    class Node implements Comparable<Node>
    {
        private char[][] board;
        private Integer score;
        private Point move;
        
        /**
         * Constructor
         */
        Node(char[][] board, Point move)
        {
            this.board = board;
            this.move = move;
        }
    
        /**
         * Sets score
         */
        public void setScore(int score)
        {
            this.score = score;
        }
        
        /**
         * Returns board configuration
         */
        public char[][] getBoard()
        {
            return board;
        }
        
        /**
         * Returns score
         */
        public int getScore()
        {
            return score;
        }
        
        /**
         * Returns the move which resulted in this configuration
         */
        public Point getMove()
        {
            return move;
        }
        
        /**
         * CompareTo used for determining min/max scores
         */
        public int compareTo(Node other){
            return score.compareTo(other.score);
        }   
    }
    
}//end of Game class


